#include "sensor.h"

#include <QString>
#include <QByteArray>
#include <QTextStream>
#include <QMessageBox>
#include <QFile>
#include <QDir>
#include <QVariant>
#include <QtDebug>
#include <QTimer>
#include <QRegExp>

#ifdef Q_WS_WIN
#define _WIN32_DCOM
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <ole2.h>
#include <Wbemidl.h>
#endif

Sensor::Sensor(QObject *parent, int timeout)
    : QObject(parent), m_tripPoint(-1.0), m_min(0.0), m_max(10.0), m_interval(5.0),
      m_value(5.0), m_buddy(0)
{
    QTimer *timer = new QTimer(this);

    // Ensure first update happens ASAP.
    QTimer::singleShot(0, this, SLOT(updateValue()));

    connect(timer, SIGNAL(timeout()), SLOT(updateValue()));
    timer->start(timeout);
}

double Sensor::value() const
{
    return m_value;
}

void Sensor::setBuddy(QObject *buddy)
{
    if(m_buddy)
	disconnect(m_buddy);

    m_buddy = buddy;

    if(!m_buddy)
	return;

    connect(this, SIGNAL(valueChanged(double)), buddy, SLOT(setValue(double)));

    m_buddy->setProperty("minimum", m_min);
    m_buddy->setProperty("maximum", m_max);
    m_buddy->setProperty("label", m_label);
    m_buddy->setProperty("tickInterval", m_interval);
    m_buddy->setProperty("tripPoint", m_tripPoint);
    m_buddy->setProperty("value", m_min);

    emit valueChanged(m_value);
}

void Sensor::setRange(double min, double max)
{
    m_min = min;
    m_max = max;

    if(!m_buddy)
	return;

    m_buddy->setProperty("minimum", m_min);
    m_buddy->setProperty("maximum", m_max);
}

void Sensor::setTripPoint(double tripPoint)
{
    m_tripPoint = tripPoint;

    if(m_buddy)
	m_buddy->setProperty("tripPoint", tripPoint);
}

void Sensor::setInterval(double interval)
{
    m_interval = interval;

    if(m_buddy)
	m_buddy->setProperty("tickInterval", interval);
}

void Sensor::setLabel(const QString &label)
{
    m_label = label;

    if(m_buddy)
	m_buddy->setProperty("label", label);
}

void Sensor::setValue(double value)
{
    if(value != m_value) {
	m_value = value;
	emit valueChanged(m_value);
    }
}

int Sensor::readValueFromFile(const QString &fileName, bool *ok)
{
    QFile f(fileName);

    if(!f.open(QIODevice::ReadOnly)) {
	qDebug() << "Unable to open file" << fileName << endl;
	if(ok)
	    *ok = false;
	return 0;
    }

    QTextStream ts(&f);

    QString l = ts.readLine(); // Read in value.
    int val = l.simplified().toInt(ok);

    return val;
}

int Sensor::readI2CValue(const QString &fileName, bool *ok)
{
    if(m_fileName.isEmpty() || !QFile::exists(m_fileName)) {
	QDir sysI2C("/sys/bus/platform/devices/");

	// Find list of sensors.
	QStringList I2CDevices = sysI2C.entryList(QDir::Dirs | QDir::NoDotAndDotDot);
	if(I2CDevices.isEmpty()) {
	    if(ok)
		*ok = false;
	    return 0; // No devices, nothing I can do for now.
	}

	foreach(QString dirName, I2CDevices) {
	    if(!dirName.startsWith("it87"))
		continue;

	    m_fileName = sysI2C.absoluteFilePath(dirName)
		       + QDir::separator()
		       + fileName;
	    break;
	}
    }

    return readValueFromFile(m_fileName, ok);
}

#ifdef Q_WS_X11
FanSpeedSensor::FanSpeedSensor(QObject *parent)
  : Sensor(parent, 100)
{
    setLabel("RPM");
    setRange(0.0, 3000.0);
    setTripPoint(2500.0);
    setInterval(300.0);
}

void FanSpeedSensor::updateValue()
{
    bool ok = true;
    int val = readI2CValue("fan1_input", &ok);

    if(!ok)
	return;

    setValue(val);
}

CPUTempSensor::CPUTempSensor(QObject *parent, int CPUID)
  : Sensor(parent, 2000), m_cpuId(CPUID + 1)
{
    setTripPoint(45.0);
    setRange(25.0, 60.0);
    setInterval(5.0);
    setLabel(QString::fromUtf8("CPU (°C)"));
}

void CPUTempSensor::updateValue()
{
    // Just pick the first device.  I assume that temp1_input is the one we
    // want.  A real GUI would need scaling factors and offsets and a way to
    // guess the right temperature because lm_sensors is crap.
    bool ok = true;
    int val = readI2CValue(QString("temp%1_input").arg(m_cpuId), &ok);

    if(!ok)
	return;

    setValue(val / 1000.0);
}

NetSpeedSensor::NetSpeedSensor(QObject *parent, NetSpeedType speed)
  : Sensor(parent, 250), m_speed(speed)
{
    if(speed == RxSpeed) {
	setLabel("RX (KBps)");
	setRange(50.0, 500.0);
	setInterval(50.0);
    }
    else {
	setLabel("TX (KBps)");
	setRange(5.0, 50.0);
	setInterval(5.0);
    }

    m_bytesRx = currentRxBytes();
    m_timeElapsed.start();
}

void NetSpeedSensor::updateValue()
{
    quint64 cur = currentRxBytes();
    quint64 delta = cur - m_bytesRx;
    m_bytesRx = cur;

    // We have delta, find time elapsed.
    int msecsElapsed = m_timeElapsed.restart();

    if(msecsElapsed == 0) {
	qDebug("No time elapsed for measurement.");
	return;
    }

    double KBps = (delta * 1000.0 / 1024.0) / msecsElapsed;
    setValue(KBps);
}

quint64 NetSpeedSensor::currentRxBytes()
{
    QFile f("/proc/net/dev");

    if(!f.open(QIODevice::ReadOnly))
        return 0;

    QTextStream ts(&f);

    ts.readLine();
    ts.readLine(); // Skip first 2 lines, which is header information.

    while(!ts.atEnd()) {
        QString data = ts.readLine();
        QRegExp nameRE("^([\\sa-zA-Z:0-9]+):(.*)$");

        if(nameRE.indexIn(data) == -1) {
            qDebug("Wrong name?");
            return 0;
        }

        // Change this to the device you'd like to query.
        if(nameRE.cap(1).simplified() != "eth0")
            continue;

        QStringList dataParts = nameRE.cap(2).split(QRegExp("\\s+"));
        if(!dataParts.isEmpty()) {
	    if(m_speed == RxSpeed)
		return dataParts[0].toULongLong();
	    else
		return dataParts[8].toULongLong();
        }
        else {
            qDebug("No data with line.");
            return 0;
        }
    }

    return 0;
}

UPCSensor::UPCSensor(QObject *parent, const QString &parameter)
  : Sensor(parent, 5000), m_parameter(parameter)
{
}

void UPCSensor::updateValue()
{
//    setValue(value());
}

CPUFreqSensor::CPUFreqSensor(QObject *parent, int cpuid)
  : Sensor(parent, 50), m_cpuid(cpuid)
{
    // Get max and min frequencies.
    bool ok = false;
    QString file = QString("/sys/devices/system/cpu/cpu%1/cpufreq/cpuinfo_max_freq").arg(cpuid);

    int maxFreq = readValueFromFile(file, &ok);
    int minFreq = 0;

    if(ok) {
	file = QString("/sys/devices/system/cpu/cpu%1/cpufreq/cpuinfo_min_freq").arg(cpuid);
	minFreq = readValueFromFile(file, &ok);
    }

    setLabel("CPU Frequency (MHz)");
    if(ok) {
	setRange(minFreq / 1000, maxFreq / 1000);
	setInterval(100);

	updateValue();
    }
}

void CPUFreqSensor::updateValue()
{
    bool ok = false;
    // cpuinfo_cur_freq is not readable unless root.
    QString file = QString("/sys/devices/system/cpu/cpu%1/cpufreq/scaling_cur_freq").arg(m_cpuid);

    int value = readValueFromFile(file, &ok);
    if(ok)
	setValue(value / 1000);
}

CPULoadSensor::CPULoadSensor(QObject *parent, int cpuId)
  : Sensor(parent, 950), m_cpuId(cpuId)
{
    setRange(20.0, 100.0);
    setLabel("CPU Load (%)");
    setInterval(10.0);

    readCPULoad(m_lastLoad);
}

// Code below adapted from ksysguard, written by Chris Schlaeger (GPLv2).
void CPULoadSensor::updateValue()
{
    CPULoadInfo cur;

    if(!readCPULoad(cur))
	return;

    quint64 deltaTicks = cur.userTicks - m_lastLoad.userTicks +
                         cur.sysTicks - m_lastLoad.sysTicks +
			 cur.niceTicks - m_lastLoad.niceTicks +
			 cur.idleTicks - m_lastLoad.idleTicks;

    double totalLoad = 100.0 - 100.0 * (cur.idleTicks - m_lastLoad.idleTicks)
				   / deltaTicks;
    m_lastLoad = cur;

    totalLoad = qBound(0.0, totalLoad, 100.0);
    setValue(totalLoad);
}

bool CPULoadSensor::readCPULoad(CPULoadInfo &load)
{
    QFile stats("/proc/stat");

    if(!stats.open(QIODevice::ReadOnly))
	return false;

    QByteArray cpuId("cpu");

    if(m_cpuId != -1)
	cpuId.append(QByteArray::number(m_cpuId));
    else
	cpuId.append(' '); // Don't inadvertently start with e.g. cpu0

    QByteArray cpuInfo;
    bool found = false;

    forever {
	cpuInfo = stats.readLine().simplified();
	if(cpuInfo.isEmpty())
	    break;

	int sep = cpuInfo.indexOf(' ');
	if(sep == -1 || !cpuInfo.startsWith(cpuId))
	    continue;

	// In right string, break out info entries
	int start = sep + 1;
	sep = cpuInfo.indexOf(' ', start);
	if(sep == -1) {
	    qWarning() << "Stuff not found in str, starting from " << start << cpuInfo << cpuInfo.mid(start);
	    break;
	}
	load.userTicks = cpuInfo.mid(start, sep - start).toULongLong();

	start = sep + 1;
	sep = cpuInfo.indexOf(' ', start);
	if(sep == -1)
	    break;
	load.sysTicks = cpuInfo.mid(start, sep - start).toULongLong();

	start = sep + 1;
	sep = cpuInfo.indexOf(' ', start);
	if(sep == -1)
	    break;
	load.niceTicks = cpuInfo.mid(start, sep - start).toULongLong();

	start = sep + 1;
	sep = cpuInfo.indexOf(' ', start);
	if(sep == -1)
	    break;
	load.idleTicks = cpuInfo.mid(start, sep - start).toULongLong();

	found = true;
	break;
    }

    if(!found) {
	// Can't parse the string.
	return false;
    }

    return true;
}
#endif

#ifdef Q_WS_WIN
CPULoadSensor::CPULoadSensor(QObject *parent)
  : Sensor(parent, 250)
{
    setRange(0.0, 100.0);
    setLabel("CPU Load (%)");
    setInterval(10.0);

    HRESULT hres;

//    hres = CoInitialize(0);
    CoUninitialize();
    hres = CoInitializeEx(0, COINIT_MULTITHREADED);

    if(FAILED(hres)) {
	qDebug() << "Unable to initialize COM";
	QMessageBox::warning(0, "Exception thrown", "Unable to initialize COM");
        return;
    }

    hres = CoInitializeSecurity(
        NULL,
	-1,
	NULL,
	NULL,
	RPC_C_AUTHN_LEVEL_DEFAULT,
	RPC_C_IMP_LEVEL_IMPERSONATE,
	NULL,
	EOAC_NONE,
	NULL
    );

    if(FAILED(hres)) {
	CoUninitialize();
	QMessageBox::warning(0, "Exception thrown", "COM security still sucks.");
    }
}

CPULoadSensor::~CPULoadSensor()
{
    CoUninitialize();
}

template<class T>
class CComPtr
{
public:
    CComPtr() : m_ptr(0) { }
    ~CComPtr() { if(m_ptr) m_ptr->Release(); m_ptr = 0; }

    T** operator&() { return &m_ptr; }

    T* operator->() { return m_ptr; }

private:
    T *m_ptr;
};

void CPULoadSensor::updateValue()
{
    CComPtr<IWbemLocator> pLoc;

    HRESULT hres = CoCreateInstance(
        CLSID_WbemLocator,
        0,
        CLSCTX_INPROC_SERVER,
        IID_IWbemLocator, (LPVOID*) &pLoc);

    if(FAILED(hres)) {
        QMessageBox::warning(0, "Exception thrown", "WBEM is gay");
        return;
    }

    CComPtr<IWbemServices> pSvc;
    hres = pLoc->ConnectServer(L"root\\cimv2", NULL, NULL, NULL,
              0, NULL, NULL, &pSvc);

    if(FAILED(hres)) {
        QMessageBox::warning(0, "Exception thrown", "Could not connect to server.");
        return;
    }

    CComPtr<IEnumWbemClassObject> enumerator;
    hres = pSvc->ExecQuery(L"WQL", L"SELECT * FROM Win32_Processor",
                           WBEM_FLAG_FORWARD_ONLY, NULL, &enumerator);
    if(FAILED(hres)) {
        QMessageBox::warning(0, "Exception thrown", "Unable to execute query.");
        return;
    }

    ULONG retcnt;
    CComPtr<IWbemClassObject> processor;
    hres = enumerator->Next(0, 1L, &processor, &retcnt);
    if(FAILED(hres)) {
        QMessageBox::warning(0, "Exception thrown", "Unable to enumerate");
        return;
    }

    if(retcnt == 0) {
        QMessageBox::warning(0, "Exception thrown", "No results returned.");
        return;
    }

    VARIANT prop;
    hres = processor->Get(L"LoadPercentage", 0, &prop, NULL, NULL);
    if(FAILED(hres)) {
        QMessageBox::warning(0, "Exception thrown", "Seriously now, wtf.");
        return;
    }

    setValue(prop.iVal);
}
#endif

// vim: set fileencoding=utf-8 encoding=utf-8:
